/*******************************************************************************
*
* Description : Linux Device Driver to access PassMark PCIe loopback card
*
********************************************************************************/
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/init.h>
#include <linux/cdev.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <asm/current.h>
#include <linux/pci_regs.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/kdev_t.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>

#include "pcie_common.h"

//Uncomment to status messages to DMESG
//#define DEBUG_MODE

#define DRV_NAME "PM_PCIE_DRIVER"
#define PCI_DEV_NAME "PM_PCIE_DEV"
#define DRV_VERSION	"1.1001"

#define PCI_BAR_0 0	
#define PCI_BAR_1 1

#define MAX_DMA_BUFFER 8192

// Maximum number of PCIE cards we will detect and communicate with
#define MAX_NUM_PCIECARDS 10 


#define REG_IGL2_LED_CTRL 		0x00A0	/ 4 	// LED register 
#define REG_IGL2_DMA			0x8008	/ 4 	// DMA_DIR register 
#define REG_IGL2_THROUGHPUT		0x0802	/ 4 
#define REG_IGL2_SET_WRITE_MODE		0x0805	/ 4
#define REG_IGL2_ADC_CHANNEL		0x0004 	/ 4	// ADC channel selection register
#define REG_IGL2_BASE_ADDR_WR		0x8200 	/ 4 	// Host -> Device
#define REG_IGL2_BASE_ADDR_RD		0x8028 	/ 4 	// Device -> Host
#define REG_IGL2_FW_VERSION		0x0010 	/ 4 	// Firmware version
#define REG_IGL2_INT_ENABLED		0x000C 	/ 4	// Interrupts enabled/disabled
#define REG_IGL2_XFER_SIZE		0x8000	/ 4	// Transfer size ub 128B blocks

#define REG_IGL2_SER_NUM0		0x8220	/ 4 	// Write a '1' to retireve serial numberm, then read this for lowest 32bits of serial numer
#define REG_IGL2_SER_NUM1		0x8228 	/ 4	// next 32bits of serial
#define REG_IGL2_SER_NUM2		0x82A0	/ 4 	// next 32bits of serial
#define REG_IGL2_SER_NUM3		0x82A8	/ 4	// last 32bits of serial
#define REG_IGL2_DMA_CH0_STATUS		0x8100 	/ 4	//DMA channel 0 status
#define REG_IGL2_DMA_CH1_STATUS		0x8108 	/ 4	//DMA channel 1 status
#define REG_IGL2_LSRAM_REG_DATA_L 	0x0094 / 4 	//LSRAM contents of index specified in LSRAM_REG_ADDRIOCTL_IGL2_ADC_STATS
#define REG_IGL2_LSRAM_REG_DATA_H 	0x009B / 4 	//LSRAM contents of index specified in LSRAM_REG_ADDR
#define REG_IGL2_LSRAM_REG_ADDR 	0x009C / 4 	//LSRAM memory index address 
#define REG_IGL2_ADC_INT_REG		0x8010 / 4	//Request ADC stat read?
#define	REG_IGL2_ADC_STAT_A0		0x8230 / 4	//Temperature sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_STAT_B5		0x8238 / 4	//3.3V aux sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_STAT_A2		0x8250 / 4	//12V pcie sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_STAT_B2		0x8258 / 4	//3.3V pcie sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_STAT_A3		0x8260 / 4	//12V pcie (Duplicate) sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_STAT_B3		0x8268 / 4	//12V molex sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_STAT_A4		0x8270 / 4	//5V molex sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_STAT_B4		0x8278 / 4	//12V molex (Duplicate) sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_AVG_0		0x82C8 / 4	//Average of 3.3Vaux (ch B5) 31:16, Avg of temp (ch A0) 15:0
#define	REG_IGL2_ADC_AVG_2		0x82D0 / 4	//Average of 3.3V (ch B2) 31:16, Avg of 12V (ch A2) 15:0
#define	REG_IGL2_ADC_AVG_3		0x82D8 / 4	//Average of 12V molex (ch B3) 31:16, Avg of 12V (ch A3) 15:0
#define	REG_IGL2_ADC_AVG_4		0x82E0 / 4	//Average of 12V molex (ch B4) 31:16, Avg of 5V molex (ch A4) 15:0
#define REG_IGL2_ADC_MODE		0x8280 / 4	//Change between ADC and stats mode  used for resetting values
#define	REG_IGL2_LINK_WIDTH		0x0008 / 4	//Pcie link width, bit 0 = high is x4, low if x1
#define REG_IGL2_CLK_SEL		0x8340 / 4	//Clock selection, write 0x11AA0000 for 160, 0x11AA0001 for 166

//Debug registers
#define REG_IGL2_AAW_CHAN	0x82F0 / 4
#define REG_IGL2_AAW_ADDRINT	0x82F8 / 4
#define REG_IGL2_AAW_ADDR	0x8300 / 4
#define REG_IGL2_ADW_CHAN	0x8308 / 4
#define REG_IGL2_ADR_CHAN	0x8310 / 4
#define REG_IGL2_AAR_CHAN	0x8318 / 4
#define REG_IGL2_AAR_GEN	0x8320 / 4
#define REG_IGL2_AAR_ADDRINT	0x8328 / 4

//Clock options
#define	CLOCK_120		0x11AA0000
#define CLOCK_166		0x11AA0001

//Stats mode
#define STATS_ON		0x11AA0001
#define STATS_OFF		0x11AA0000


//Register values
#define PCIE_CLEAR_DMA		0x11AA0007
#define PCIE_CLEAR_INT		0x11AA0009

//Using the same names as the windows driver
#define MODE_READ_ADC 		0x11AA0002
#define MODE_READ_LSRAM		0x11AA0006
#define MODE_WRITE_LSRAM 	0x11AA0005
#define MODE_WRITE_LOOPBACK	0x11AA0008

//ADC channel options
#define PCIE_ADC_12V_PCIE_BUF_AND_3V3_PCIE_BUF 	0
#define PCIE_ADC_12V_PCIE_BUF_AND_TEMP_SENS_BUF 1
#define PCIE_ADC_12V_PCIE_AND_3V3_PCIE		2
#define PCIE_ADC_12V_PCIE_AND_5V_MOLEX		3
#define PCIE_ADC_12V_MOLEX_AND_5V_MOLEX		4
#define PCIE_ADC_NUM_OPTIONS			5

#define SUCCESS 0
#define ENABLE 1
#define DISABLE 0



// channel status registers
typedef struct _DMA_CSR_
{
	unsigned int Count : 16 ; 	//Bit 0 -15
	unsigned int Reserved : 12 ; 	//Bit 16 - 27
	unsigned int IntFlag : 1 ; 	//Bit 28
	unsigned int Interrupt : 1 ; 	//Bit 29
	unsigned int Loopback : 1 ; 	//Bit 30 Only in DMA CSR ch 1
	unsigned int Done : 1; 		//Bit 31
}
DMA_CSR;

typedef struct _LINK_CAPS_
{
 	unsigned int MaxLinkSpeed : 4 ; 	//Bit 0 - 3
	unsigned int MaxLinkWidth : 6 ; 	//Bit 4 - 9
	unsigned int ASPM : 2 ; 	//Bit 10-11
	unsigned int L0ExitLatency : 3 ; 	//Bit 12-14
	unsigned int L1ExitLatency : 3 ; 	//Bit 15-17
	unsigned int LinkSpeedsVector : 7; 		//Bit 18-24
}
LINK_CAPS; 

typedef struct _PCIE_CAPS
{
 	unsigned int CapVersion : 4 ; 	//Bit 0 - 3
	unsigned int Type : 4; 	//Bit 4 - 7
	unsigned int Slot : 1 ; 	//Bit 8
	unsigned int IntMsgNumber : 5 ; 	//Bit 9-13
	unsigned int RSVD : 2 ; 	//Bit 14-15
}
PCIE_CAPS;

typedef struct _LINK_STATUS_
{
 	unsigned int CurLinkSpeed : 4 ; 	//Bit 0 - 3
	unsigned int CurLinkWidth : 6 ; 	//Bit 4 - 9
	unsigned int Undef : 1 ; 	//Bit 10
	unsigned int LinkTrain : 1 ; 	//Bit 11
	unsigned int SlotClock : 1; 	//Bit 12
	unsigned int DataLink : 1; 	//Bit 13
	unsigned int LBMS : 1; 	//Bit 14
	unsigned int LABS : 1; 	//Bit 15
}
LINK_STATUS ;


union 
{
	DMA_CSR bits;
	unsigned long ulong;
} DMA0Csr;

union 
{
	DMA_CSR bits;
	unsigned long ulong;
} DMA1Csr;


/* Structure to hold driver specific resources */
struct pm_pcie_dev {
	u32 bar0_base;
	u32 bar0_size;
	volatile void __iomem *base_addr0; 	
	u32 bar1_base;
	u32 bar1_size;
	volatile void __iomem *base_addr1; 	
	dma_addr_t bus_addr_wr; 		
	unsigned char* addr_wr;		

	dma_addr_t bus_addr_rd;			
	unsigned char* 	addr_rd;		
	
	dev_t pci_dev_num; 			
	struct pci_dev *pdev;
	struct cdev *pci_cdev;
 	struct device* device;

	unsigned int wr_buf[32]; 
	unsigned int rd_buf[32]; 
	unsigned int wr_buf_tmp[32];

	CONFIG_INFO currentConfig; 		
	wait_queue_head_t waitQueue; 
	bool interruptFinished ;
};



static unsigned long intCounter;
static unsigned int numPCICards = 0;
static unsigned int pm_major_id = 0;

static struct class *pm_pcie_class = NULL;

	
struct dev_ext *pdevExt; 
static struct pm_pcie_dev * pm_pcie_cards = NULL; 

static struct pci_device_id ms_pci_tbl[] = {
	{ 0x11AA, 0x11AA, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{0,}
}; 

MODULE_DEVICE_TABLE(pci, ms_pci_tbl);


void PrintDebugVals(unsigned int *ptr1)
{
	DEBUG_REG_VALS debugVals;
	memset(&debugVals, 0, sizeof(debugVals));

	debugVals.AAW_CHAN = ioread32(ptr1 + REG_IGL2_AAW_CHAN); 
	debugVals.AAW_ADDRINT = ioread32(ptr1 + REG_IGL2_AAW_ADDRINT);
	debugVals.AAW_ADDR = ioread32(ptr1 + REG_IGL2_AAW_ADDR);
	debugVals.ADW_CHAN = ioread32(ptr1 + REG_IGL2_ADW_CHAN);
	debugVals.ADR_CHAN = ioread32(ptr1 + REG_IGL2_ADR_CHAN);
	debugVals.AAR_CHAN = ioread32(ptr1 + REG_IGL2_AAR_CHAN);
	debugVals.AAR_GEN = ioread32(ptr1 + REG_IGL2_AAR_GEN);
	debugVals.AAR_ADDRINT = ioread32(ptr1 + REG_IGL2_AAR_ADDRINT);

	pr_info("Debug Values:\nAAW_CHAN: %x\nAAW_ADDRINT: %x\nAAW_ADDR: %x\nADW_CHAN: %x\nADR_CHAN: %x\nAAR_CHAN: %x\nAAR_GEN: %x\nAAR_ADDRINT: %x\n",
	debugVals.AAW_CHAN, debugVals.AAW_ADDRINT,  debugVals.AAW_ADDR,  debugVals.ADW_CHAN, 
	debugVals.ADR_CHAN, debugVals.AAR_CHAN, debugVals.AAR_GEN, debugVals.AAR_ADDRINT); 

}
/**
* ms_pci_interrupt() - PCIe interrupt Handler
*
*  @irq : PCIe interrupt number
*/
static irqreturn_t ms_pci_interrupt(int irq, void* dev_id, struct pt_regs * regs)
{
	#ifdef DEBUG_MODE
	pr_info("%s: Got interrupt\n", DRV_NAME);
	#endif

	struct pm_pcie_dev* pm_pcie_card = (struct pm_pcie_dev*) dev_id;

	{
			
		unsigned int *ptr1 = (unsigned int*)pm_pcie_card->base_addr1;
		DMA0Csr.ulong = ioread32((ptr1 + REG_IGL2_DMA_CH0_STATUS));
		DMA1Csr.ulong = ioread32((ptr1 + REG_IGL2_DMA_CH1_STATUS));

		if( DMA0Csr.bits.Interrupt || DMA1Csr.bits.Interrupt)
		{
			unsigned int* ptr1 = (unsigned int *)(pm_pcie_card->base_addr1);

			#ifdef DEBUG_MODE
			pr_info("%s: Interrupt intCounter = %ld  DMA0Csr %x DMA1Csr %x\n", DRV_NAME, intCounter, DMA0Csr.ulong, DMA1Csr.ulong);

			pm_pcie_card->wr_buf_tmp[0] = 0;
			iowrite32(pm_pcie_card->wr_buf_tmp[0], (ptr1 + REG_IGL2_LSRAM_REG_ADDR)); 
			unsigned intLS1 = ioread32((ptr1 + REG_IGL2_LSRAM_REG_DATA_L));
			unsigned intLS2 = ioread32((ptr1 + REG_IGL2_LSRAM_REG_DATA_H));

			pr_info("%s: Interrupt LSRAM %x %x\n", DRV_NAME, intLS1, intLS2 );
			#endif

						pm_pcie_card->wr_buf_tmp[0] = PCIE_CLEAR_INT;
			iowrite32(pm_pcie_card->wr_buf_tmp[0], (ptr1 + REG_IGL2_DMA)); 

			intCounter = intCounter + 1; 

			pm_pcie_card->interruptFinished = 1;
			wake_up_interruptible(&pm_pcie_card->waitQueue);

			#ifdef DEBUG_MODE
			pr_info("%s: returned IRQ_HANDLED\n", DRV_NAME);
			#endif

			return IRQ_HANDLED;
		}
		else
			return IRQ_NONE;
	}

		return IRQ_NONE;
}   



/**
* ms_pci_remove() - Uninitialize the PCIe device
*
*  @pdev : Platform Device
*/
static void ms_pci_remove(struct pci_dev *pdev)
{
				
	printk(KERN_INFO "%s: ms_pci_remove\n", DRV_NAME);
}


   
/**
* char_dev_open() - Open Char Device
*
*  @inode : Device Driver Information
*  @file  : Pointer to the file struct
*/
static int char_dev_open(struct inode *inode,
			    struct file  *file)
{
	unsigned int major = imajor(inode);
	unsigned int minor = iminor(inode);	
	struct pm_pcie_dev * dev = NULL;

	printk(KERN_INFO "%s: open operation invoked \n", DRV_NAME);
	
	if(major != pm_major_id || minor < 0 || minor >= numPCICards)
	{
		printk(KERN_INFO "%s: Open: No device found with major %d minor %d\n", DRV_NAME, major, minor);
		return -ENODEV;
	}

		dev = &pm_pcie_cards[minor];
	file->private_data = dev;

	if(inode->i_cdev != dev->pci_cdev)
	{
		printk(KERN_INFO "%s: Open: Internal error, cdev mismatch\n", DRV_NAME);
		return -ENODEV;
	}

	
	return SUCCESS;
}

/**
* char_dev_release() - Release Char Device
*
*  @inode : Device Driver Information
*  @file  : Pointer to the file struct
*/

static int char_dev_release(struct inode *inode,
		            struct file *file)
{
	return SUCCESS;
}

static void clear_DMA_status(unsigned int *ptr1)
{
		#ifdef DEBUG_MODE
	pr_info("%s: Clear DMA status\n", DRV_NAME);
	#endif

	iowrite32(PCIE_CLEAR_DMA, (ptr1 + REG_IGL2_DMA)); 

	#ifdef DEBUG_MODE
	pr_info("%s: Clear DMA status: wrote to %p\n", DRV_NAME, (ptr1 + REG_IGL2_DMA));
	#endif
}


/**
* char_dev_read() - Copy from PCIe memory to device file /dev/PM_PCI_DEV for user read
*
*  @file : Pointer to the file struct
*  @buf  : read buffer
*  @lbuf : size of the read buffer
*  @ppos : Position of buffer

*/

static ssize_t char_dev_read(struct file *file, 
                                     char *buf,
                                     size_t lbuf,
                                     loff_t *ppos)
{

		struct pm_pcie_dev* res = (struct pm_pcie_dev*) file->private_data;
	int* ptr1 = NULL;
	int nbytes = 0; 

	if(buf == NULL)
	{
		pr_info("%s: NULL buffer passed\n", DRV_NAME);
		return -1;
	}

	if(lbuf < res->currentConfig.xferBufferSize)
	{
		pr_info("%s: User read buffer too small (%zu v %d)\n", DRV_NAME, lbuf, res->currentConfig.xferBufferSize);
		return -1;
	}

	ptr1 = (unsigned int *)(res->base_addr1);

	if(res->currentConfig.writeMode != MODE_WRITE_LOOPBACK)
	{
		res->interruptFinished = 0;
		clear_DMA_status(ptr1);
	
			iowrite32(res->currentConfig.readMode, (ptr1 + REG_IGL2_DMA)); 
	
			wait_event_interruptible(res->waitQueue, res->interruptFinished != 0);

	}

		nbytes = copy_to_user(buf, res->addr_rd, res->currentConfig.xferBufferSize);

	if( nbytes != 0 )
		pr_info("\n%s: ERROR nbytes %d\n", DRV_NAME, nbytes);

	return (res->currentConfig.xferBufferSize - nbytes);
}

/**
* char_dev_write() - Copy buffer from user and write to PCIe driver
*
*  @file : Pointer to the file struct
*  @buf  : Write buffer
*  @lbuf : size of the buffer write
*  @ppos : Position of buffer
*/


/**
 * 
 * @param inode 
 * @param filp 
 * @param cmd 
 * @param arg 
 * @return 
 */

static long dev_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
		struct pm_pcie_dev* res = (struct pm_pcie_dev*) filp->private_data;
	
	int curval = 0;
	int nbytes = 0;
	unsigned int *ptr0;
	unsigned int *ptr1;

  ptr0 = (unsigned int *)(res->base_addr0);
	ptr1 = (unsigned int *)(res->base_addr1);

	#ifdef DEBUG_MODE
	pr_info("\n%s: ptr0 = %p\n", DRV_NAME, ptr0); 
	pr_info("\n%s: ptr1 = %p\n", DRV_NAME, ptr1);
	#endif
    

	switch (cmd)
	{

		case IOCTL_IGL2_GETCB:
		{
			#ifdef DEBUG_MODE
			pr_info("%s: IOCTL_IGL2_GETCB\n", DRV_NAME);	
			#endif

			nbytes = copy_to_user(&arg, res->addr_rd, res->currentConfig.xferBufferSize);
			pr_info("\n%s: read %d of %d\n", DRV_NAME, res->currentConfig.xferBufferSize - nbytes, res->currentConfig.xferBufferSize);
		}
		break;

		case IOCTL_IGL2_SET_CLOCK:
		{
			pr_info("%s: IOCTL_IGL2_SET_CLOCK\n", DRV_NAME);
			switch(arg)
			{
				case CLOCK_LOW:
				pr_info("%s: Set clock low\n", DRV_NAME);
				iowrite32(CLOCK_120, (ptr1 + REG_IGL2_CLK_SEL));
				
				//Sleep to allow clock change to go into effect	
				msleep(1000);
				break;
	
				case CLOCK_HIGH:
				{
					pr_info("%s: Set clock high\n", DRV_NAME);
					curval = ioread32((ptr1 + REG_IGL2_CLK_SEL));

					if(res->currentConfig.pcieVer < 2)
					{
						pr_info("%s: Error IOCTL_IGL2_SET_CLOCK PCIe version 1 detected, cannot set clock to high speed\n", DRV_NAME);	
						return -1;			
					}
					
					//Check if clock is aleady high and only set when low
					if(curval == 0)
					{
						iowrite32(CLOCK_166, (ptr1 + REG_IGL2_CLK_SEL));
						//Sleep to allow clock change to go into effect	
						msleep(1000);
					}
					curval = ioread32((ptr1 + REG_IGL2_CLK_SEL));
					pr_info("%s: REG_IGL2_CLK_SEL %x\n", DRV_NAME, curval);

	
				break;
				}
			}
		}
		break;

		case IOCTL_IGL2_LEDCTL: /* LED Access */
		{
			#ifdef DEBUG_MODE
			pr_info("%s: IOCTL_IGL2_LEDCTL\n", DRV_NAME);
			#endif

			iowrite32(arg, (ptr1 + REG_IGL2_LED_CTRL)); 
			
		}
		break;

		case IOCTL_IGL2_CLEAR_DMA_REG:
		{
			#ifdef DEBUG_MODE
			pr_info("%s: IOCTL_IGL2_CLEAR_DMA_REG\n", DRV_NAME);
			#endif
		 clear_DMA_status(ptr1);
		}
		break;

		case IOCTL_IGL2_RESET_STATS_REG:
		{
			#ifdef DEBUG_MODE
			pr_info("%s: IOCTL_IGL2_RESET_STATS_REG\n", DRV_NAME);
			#endif

			iowrite32(STATS_OFF, (ptr1 + REG_IGL2_ADC_MODE));  
			wmb();
			#ifdef DEBUG_MODE
			int tmp = ioread32((ptr1 + REG_IGL2_ADC_MODE));
			pr_info("%s: REG_IGL2_ADC_MODE: 0x%x\n", DRV_NAME, tmp);
			#endif
			
			iowrite32(STATS_ON, (ptr1 + REG_IGL2_ADC_MODE)); 
			wmb();
			#ifdef DEBUG_MODE
			tmp = ioread32((ptr1 + REG_IGL2_ADC_MODE));
			pr_info("%s: REG_IGL2_ADC_MODE: 0x%x\n", DRV_NAME, tmp);
			#endif
		}
		break;

		case IOCTL_IGL2_SET_XFER_SIZE:
		{
			#ifdef DEBUG_MODE
			pr_info("%s: REG_IGL2_XFER_SIZE\n", DRV_NAME);
			#endif
			if(arg <= 64)
			{
 				iowrite32(arg, (ptr1 + REG_IGL2_XFER_SIZE)); 

				res->currentConfig.xferSize = arg;
				res->currentConfig.xferBufferSize = arg * 128;
				
			}
		}
		break;
	
		case IOCTL_IGL2_GET_CONFIG:
		{
			#ifdef DEBUG_MODE
			pr_info("%s: IOCTL_IGL2_GET_CONFIG\n", DRV_NAME);
			#endif
			CONFIG_INFO configBuf;
			memset(&configBuf, 0, sizeof(configBuf));

								
			configBuf.linkWidth = ioread32( (ptr1 + REG_IGL2_LINK_WIDTH));
			configBuf.adcCh = res->currentConfig.adcCh;
			configBuf.xferSize = ioread32(ptr1 + REG_IGL2_XFER_SIZE);
			configBuf.xferBufferSize = configBuf.xferSize * 128;
			configBuf.readMode =  res->currentConfig.readMode;
			configBuf.writeMode =  res->currentConfig.writeMode;
			configBuf.ledStatus = ioread32((ptr1 + REG_IGL2_LED_CTRL));
			configBuf.sn0 = ioread32((ptr1 + REG_IGL2_SER_NUM0));
			configBuf.sn1 = ioread32(ptr1 + REG_IGL2_SER_NUM1);
			configBuf.sn2 = ioread32((ptr1 + REG_IGL2_SER_NUM2));
			configBuf.sn3 = ioread32((ptr1 + REG_IGL2_SER_NUM3));
			configBuf.fwVersion = ioread32((ptr1 + REG_IGL2_FW_VERSION));
			configBuf.pcieVer = res->currentConfig.pcieVer;

			if(copy_to_user((CONFIG_INFO*) arg, &configBuf, sizeof(configBuf)) != 0)
			{
				pr_info("%s: Error IOCTL_IGL2_GET_CONFIG copy_to_user failed\n", DRV_NAME);	
				return -1;		
			}
		}
		break;

		case IOCTL_IGL2_SET_READ_MODE:
		{
			#ifdef DEBUG_MODE
			pr_info("%s: IOCTL_IGL2_SET_READ_MODE from %d to %d\n", DRV_NAME, res->currentConfig.readMode, arg);
			#endif

			switch(arg)
			{
				case PCIE_READ_ADC_MODE:
					res->currentConfig.readMode  = MODE_READ_ADC;
					break;
			
				case PCIE_READ_LSRAM_MODE:
					res->currentConfig.readMode  = MODE_READ_LSRAM;
					break;
			}

		}
		break;

	
		case IOCTL_IGL2_SET_WRITEMODE:
		{
			#ifdef DEBUG_MODE
			pr_info("%s: IOCTL_IGL2_SET_WRITEMODE from %d to %d\n", DRV_NAME, res->currentConfig.writeMode, arg);
			#endif

			switch(arg)
			{
				case PCIE_WRITE_BENCHMARK_MODE:
					res->currentConfig.writeMode = MODE_WRITE_LSRAM;			
					break;
			
				case PCIE_WRITE_LOOPBACK_MODE:
					res->currentConfig.writeMode = MODE_WRITE_LOOPBACK;
					break;
			}
		}
	break;

	case IOCTL_IGL2_GET_DEBUG_VALS:
	{

		DEBUG_REG_VALS debugVals;
		memset(&debugVals, 0, sizeof(debugVals));

		debugVals.AAW_CHAN = ioread32(ptr1 + REG_IGL2_AAW_CHAN); 
		debugVals.AAW_ADDRINT = ioread32(ptr1 + REG_IGL2_AAW_ADDRINT);
		debugVals.AAW_ADDR = ioread32(ptr1 + REG_IGL2_AAW_ADDR);
		debugVals.ADW_CHAN = ioread32(ptr1 + REG_IGL2_ADW_CHAN);
		debugVals.ADR_CHAN = ioread32(ptr1 + REG_IGL2_ADR_CHAN);
		debugVals.AAR_CHAN = ioread32(ptr1 + REG_IGL2_AAR_CHAN);
		debugVals.AAR_GEN = ioread32(ptr1 + REG_IGL2_AAR_GEN);
		debugVals.AAR_ADDRINT = ioread32(ptr1 + REG_IGL2_AAR_ADDRINT);

		if(copy_to_user((unsigned int*) arg, &debugVals	, sizeof(debugVals)) != 0)
		{
			pr_info("%s: Error IOCTL_IGL2_GET_DEBUG_VALS copy_to_user failed\n", DRV_NAME);	
			return -1;		
		}
	}
	break;


	
	case IOCTL_IGL2_SET_ADC_CH:
	{
		#ifdef DEBUG_MODE
		pr_info("%s: IOCTL_IGL2_SET_ADC_CH\n", DRV_NAME);
		#endif

				#ifdef DEBUG_MODE
		pr_info("%s: Set ADC channel\n", DRV_NAME);
		#endif

		iowrite32(arg, (ptr1 + REG_IGL2_ADC_CHANNEL)); 

		#ifdef DEBUG_MODE
		pr_info("%s: wrote 0x%x to %p\n", DRV_NAME, arg, (ptr1+REG_IGL2_ADC_CHANNEL)); 
		#endif
			
			
	}
	break;

	case IOCTL_IGL2_GET_ADC_STATS:
	{
		int tmp = 0;
		#ifdef DEBUG_MODE
		pr_info("%s: IOCTL_IGL2_GET_ADC_STATS\n", DRV_NAME);
		#endif

		ADC_STATS_RAW stats;
		memset(&stats, 0, sizeof(stats));

		tmp = ioread32(ptr1 + REG_IGL2_ADC_INT_REG);
		
		#ifdef DEBUG_MODE
		pr_info("%s: REG_IGL2_ADC_INT_REG 0x%x\n", DRV_NAME, tmp);

		tmp = ioread32(ptr1 + REG_IGL2_ADC_MODE);
		pr_info("%s: REG_IGL2_ADC_MODE 0x%x\n", DRV_NAME, tmp);
		#endif

		stats.Stat_therm = ioread32(ptr1 + REG_IGL2_ADC_STAT_A0);
		stats.Stat_3v3aux = ioread32(ptr1 + REG_IGL2_ADC_STAT_B5);
		stats.Stat_12v = ioread32(ptr1 + REG_IGL2_ADC_STAT_A2);
		stats.Stat_3v3 = ioread32(ptr1 + REG_IGL2_ADC_STAT_B2);
		stats.Stat_12vpsu = ioread32(ptr1 + REG_IGL2_ADC_STAT_B3);
		stats.Stat_5vpsu = ioread32(ptr1 + REG_IGL2_ADC_STAT_A4);

		stats.Avg_1 = ioread32(ptr1 + REG_IGL2_ADC_AVG_0);
		stats.Avg_2 = ioread32(ptr1 + REG_IGL2_ADC_AVG_2);
		stats.Avg_3 = ioread32(ptr1 + REG_IGL2_ADC_AVG_3);
		stats.Avg_4 = ioread32(ptr1 + REG_IGL2_ADC_AVG_4);

		#ifdef DEBUG_MODE
		pr_info("%s: IOCTL_IGL2_GET_ADC_STATS\n", DRV_NAME);
		pr_info("%s: Stat_therm   %d\n", DRV_NAME, stats.Stat_therm);
		pr_info("%s: Stat_3v3aux  %d\n", DRV_NAME, stats.Stat_3v3aux);
		pr_info("%s: Stat_12v     %d\n", DRV_NAME, stats.Stat_12v);
		pr_info("%s: Stat_3v3     %d\n", DRV_NAME, stats.Stat_3v3);
		pr_info("%s: Stat_12vpsu  %d\n", DRV_NAME, stats.Stat_12vpsu);
		pr_info("%s: Stat_5vpsu   %d\n", DRV_NAME, stats.Stat_5vpsu);
		pr_info("%s: stats.Avg_1  %d\n", DRV_NAME, stats.Avg_1);
		pr_info("%s: stats.Avg_2  %d\n", DRV_NAME, stats.Avg_2);
		pr_info("%s: stats.Avg_3  %d\n", DRV_NAME, stats.Avg_3);
		pr_info("%s: stats.Avg_4  %d\n", DRV_NAME, stats.Avg_4);
		#endif

		if(copy_to_user((unsigned int*) arg, &stats, sizeof(stats)) != 0)
		{
			pr_info("%s: Error IOCTL_IGL2_GET_ADC_STATS copy_to_user failed\n", DRV_NAME);	
			return -1;		
		}
		
	}
	break;
	
	case IOCTL_IGL2_GET_THROUGHPUT:
	{
		#ifdef DEBUG_MODE
		pr_info("%s: IOCTL_IGL2_GET_THROUGHPUT\n", DRV_NAME);
		#endif

		THROUGHPUT_DATA throughputBuf;
		memset(&throughputBuf, 0, sizeof(throughputBuf));

		throughputBuf.Ch0Thruput = ioread32((ptr1 + REG_IGL2_DMA_CH0_STATUS));	
		throughputBuf.Ch1Thruput = ioread32((ptr1 + REG_IGL2_DMA_CH1_STATUS));	

		if(copy_to_user((THROUGHPUT_DATA*) arg, &throughputBuf, sizeof(throughputBuf)) != 0)
		{
			pr_info("%s: Error IOCTL_IGL2_GET_THROUGHPUT copy_to_user failed\n", DRV_NAME);	
			return -1;		
		}
		
	}
	break;

	default:
	{
		pr_info("%s: Unknown IOCTL command: %x\n", DRV_NAME,cmd);
		return ENOTTY;
	}
	break;
	}
	return SUCCESS;
}

static ssize_t char_dev_write(struct file *file, 
                                     const char *buf,
                                     size_t lbuf,
                                     loff_t *ppos)
{
	int* ptr1 = NULL;
	int nbytes = 0; 
	#ifdef DEBUG_MODE
	pr_info("%s: char_dev_write\n", DRV_NAME);
	#endif

	struct pm_pcie_dev* res = (struct pm_pcie_dev*) file->private_data;


	if(buf == NULL)
	{
		#ifdef DEBUG_MODE
		pr_info("%s: NULL buffer passed\n", DRV_NAME);
		#endif

		return -1;
	}

	nbytes = copy_from_user(res->addr_wr,  buf, lbuf ); 
    
	#ifdef DEBUG_MODE
	int cnt = 0;
	for ( cnt = 0; cnt < 10; cnt++)
		pr_info("0x%x", res->addr_wr[cnt]); 
	#endif

	res->interruptFinished = 0;

   ptr1 = (unsigned int *)(res->base_addr1);
	clear_DMA_status(ptr1);

	#ifdef DEBUG_MODE
	pr_info("%s: Start DMA transfer (%x) from write\n", DRV_NAME, res->currentConfig.writeMode);
	#endif

	iowrite32(res->currentConfig.writeMode, (ptr1 + REG_IGL2_DMA)); 

	wait_event_interruptible(res->waitQueue, res->interruptFinished != 0);

	return (res->currentConfig.xferBufferSize - nbytes);
	
}


static struct file_operations char_dev_fops = {
	.owner = THIS_MODULE,
	.open = char_dev_open,
	.write = char_dev_write,
	.read = char_dev_read,
	.release = char_dev_release,
	.unlocked_ioctl = dev_ioctl // This replaced .ioctl in kernels > 2.6.36
};

   
/**
* ms_pci_probe() - Initialize char device Module
*
*  @pdev : Platform Device
*  @ent  : PCI Device ID List Entry 
*  Resource mapping for PCIe Device
*/


static int ms_pci_probe(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	int i = 0;
	unsigned int *ptr1 = NULL;
	unsigned int curVal = 0;
	int ret = 0;
	unsigned int serialnum[4];
	union 
	{
		LINK_STATUS bits;
		unsigned short  ushort;
	} LinkStatus;

	unsigned char capPointer = 0;
	bool finished = false;
	unsigned char capType = 0;
	int result = 0;
	int tmpLinkW = 0;
	int tmpFWVer = 0;
	
	memset(serialnum, 0, sizeof(serialnum));

	if(numPCICards > MAX_NUM_PCIECARDS)
	{
	   	pr_info("%s: max number of PCIe cards already claimed by driver\n", DRV_NAME);
		return -ENODEV;
	}


	init_waitqueue_head(&pm_pcie_cards[numPCICards].waitQueue);

        pr_info("%s: pci device %d found %p\n", DRV_NAME, numPCICards, pdev);

 	i = pci_enable_device(pdev);
	if (i)
		return i;

	pci_set_master(pdev);

	pm_pcie_cards[numPCICards].pci_dev_num = MKDEV(pm_major_id, numPCICards);
	pr_info("%s: pci_dev_num %d\n", DRV_NAME, pm_pcie_cards[numPCICards].pci_dev_num);

	pm_pcie_cards[numPCICards].bar0_base = pci_resource_start (pdev, PCI_BAR_0);
	pr_info("%s: PCI BAR #0 BASE = 0x%x\n", DRV_NAME, pm_pcie_cards[numPCICards].bar0_base);

	if (!pm_pcie_cards[numPCICards].bar0_base || ((pci_resource_flags (pdev, PCI_BAR_0) & IORESOURCE_MEM) == 0)) {
                dev_err(&pdev->dev, "%s: no I/O resource at PCI BAR #0\n", DRV_NAME);
        }
    else {
	   pm_pcie_cards[numPCICards].bar0_size = pci_resource_len (pdev, PCI_BAR_0);
       pr_info("%s: PCI BAR #0 SIZE = 0x%x\n", DRV_NAME, pm_pcie_cards[numPCICards].bar0_size);
    }

	if(request_mem_region(pm_pcie_cards[numPCICards].bar0_base, pm_pcie_cards[numPCICards].bar0_size, DRV_NAME) == NULL) {
		pr_info("%s: PCI BAR #0 Space not available\n", DRV_NAME);
	}
    else {
         pm_pcie_cards[numPCICards].base_addr0 = (void *)ioremap_nocache(pm_pcie_cards[numPCICards].bar0_base, pm_pcie_cards[numPCICards].bar0_size); 
         pr_info("%s: base_addr0 = 0x%p\n", DRV_NAME, pm_pcie_cards[numPCICards].base_addr0);
    }
	if(pm_pcie_cards[numPCICards].base_addr0 == NULL) {
		pr_info("%s: Unable to map BAR #0 address", DRV_NAME);
	}
		
        pm_pcie_cards[numPCICards].bar1_base = pci_resource_start (pdev, PCI_BAR_1);
        pr_info("%s: PCI BAR #1 BASE = 0x%x\n", DRV_NAME, pm_pcie_cards[numPCICards].bar1_base);

        if (!pm_pcie_cards[numPCICards].bar1_base || ((pci_resource_flags (pdev, PCI_BAR_1) & IORESOURCE_MEM) == 0)) {
                dev_err(&pdev->dev, "%s: no I/O resource at PCI BAR #1\n", DRV_NAME);
                goto err_out_free_res;
        }

        pm_pcie_cards[numPCICards].bar1_size = pci_resource_len (pdev, PCI_BAR_1);
        pr_info("%s: PCI BAR #1 SIZE = 0x%x\n", DRV_NAME, pm_pcie_cards[numPCICards].bar1_size);

        if(request_mem_region(pm_pcie_cards[numPCICards].bar1_base, pm_pcie_cards[numPCICards].bar1_size, DRV_NAME) == NULL) {
                pr_info("%s: PCI BAR #1 Space not available\n", DRV_NAME);
                goto err_out_free_res;
        }

        pm_pcie_cards[numPCICards].base_addr1 = (void *)ioremap_nocache(pm_pcie_cards[numPCICards].bar1_base, pm_pcie_cards[numPCICards].bar1_size);
        pr_info("%s: base_addr1 = 0x%p\n", DRV_NAME, pm_pcie_cards[numPCICards].base_addr1);
        if(pm_pcie_cards[numPCICards].base_addr1 == NULL) {
                pr_info("%s: Unable to map BAR #1 address", DRV_NAME);
                goto err_out_free_reg;
        }

    	ptr1 = (unsigned int*)pm_pcie_cards[numPCICards].base_addr1;



	curVal = 1;
	iowrite32(curVal, (ptr1+ REG_IGL2_SER_NUM0));  

	ret = pci_set_dma_mask(pdev, DMA_BIT_MASK(32));
	pr_info("%s: pci_set_dma_mask returned %d\n", DRV_NAME, ret);

	ret = pci_set_consistent_dma_mask(pdev, DMA_BIT_MASK(32));
	pr_info("%s: pci_set_consistent_dma_mask returned %d\n", DRV_NAME, ret);
	
	pm_pcie_cards[numPCICards].addr_wr = dma_alloc_coherent(NULL, MAX_DMA_BUFFER, &pm_pcie_cards[numPCICards].bus_addr_wr, GFP_KERNEL); 
	pm_pcie_cards[numPCICards].addr_rd = dma_alloc_coherent(NULL, MAX_DMA_BUFFER, &pm_pcie_cards[numPCICards].bus_addr_rd, GFP_KERNEL); 

	if(pm_pcie_cards[numPCICards].addr_wr == NULL || pm_pcie_cards[numPCICards].addr_rd == NULL)
	{
		pr_info("%s: ERROR unable to allocate memory\n", DRV_NAME);
		 goto err_out_free_res;
	}

	curVal = pm_pcie_cards[numPCICards].bus_addr_rd;	
	iowrite32(curVal, (ptr1+ REG_IGL2_BASE_ADDR_RD));  
	
	curVal = pm_pcie_cards[numPCICards].bus_addr_wr;
	iowrite32(curVal, (ptr1+ REG_IGL2_BASE_ADDR_WR));  

	curVal = 32;
	iowrite32(curVal, (ptr1+ REG_IGL2_XFER_SIZE));  

	serialnum[0] = ioread32((ptr1 + REG_IGL2_SER_NUM0));
	serialnum[1] = ioread32((ptr1 + REG_IGL2_SER_NUM1));
	serialnum[2] = ioread32((ptr1 + REG_IGL2_SER_NUM2));
	serialnum[3] = ioread32((ptr1 + REG_IGL2_SER_NUM3));

	pr_info("%s: Serial#: 0x%x %x %x %x\n",DRV_NAME, serialnum[0], serialnum[1], serialnum[2], serialnum[3]);
	
	//Firmware version sanity check, try to make sure not using Fw version 2 with 
	//driver V1
	tmpLinkW =  ioread32( (ptr1 + REG_IGL2_LINK_WIDTH));
	tmpFWVer = ioread32((ptr1 + REG_IGL2_FW_VERSION));

	pr_info("%s: Link width %d fw ver %d\n", DRV_NAME,tmpLinkW, tmpFWVer);
	if(tmpFWVer >= 0x20 || tmpLinkW > 1)
	{
		pr_info("%s: ERROR invalid firmware version or link width detected\n", DRV_NAME);	
		pr_info("%s: Possible using firmware V2 on PCIe card with the V1 driver\n", DRV_NAME);	
		pr_info("%s: Please try using the V2 driver\n", DRV_NAME);	
		return -1;	
	}
	
	if(pci_enable_msi(pdev))
	{
		pr_info("%s: Unable to enable MSI interrupt 0x%x\n",DRV_NAME, pdev->irq);
		goto err_out_free_reg;	
	}

	pr_info("%s: MSI interrupt 0x%x\n", DRV_NAME, pdev->irq);

	if(request_irq(pdev->irq, (irq_handler_t)ms_pci_interrupt, IRQF_SHARED, DRV_NAME, (void*)&pm_pcie_cards[numPCICards]) ) 
	{
		pr_info("%s: unable to register irq 0x%x\n", DRV_NAME, pdev->irq);
        	goto err_out_free_irq;
	} 
	

	pr_info("%s: Create /dev file\n", DRV_NAME);

	if (!( pm_pcie_cards[numPCICards].pci_cdev = cdev_alloc ())) 
	{
        	  printk (KERN_ERR "%s: cdev_alloc() failed\n", DRV_NAME);
            return -1;
 	}
	
	pr_info("%s: cdev_init\n", DRV_NAME);
	cdev_init( pm_pcie_cards[numPCICards].pci_cdev, &char_dev_fops);

	pr_info("%s: cdev_add\n", DRV_NAME);
	ret=cdev_add( pm_pcie_cards[numPCICards].pci_cdev, pm_pcie_cards[numPCICards].pci_dev_num, MAX_NUM_PCIECARDS);
	if( ret < 0 ) 
	{
		printk(KERN_INFO "%s: Error registering device driver\n", DRV_NAME);
	        cdev_del ( pm_pcie_cards[numPCICards].pci_cdev);
		return -1;
	}

	pr_info("%s: device_create\n", DRV_NAME);
	pm_pcie_cards[numPCICards].device = device_create (pm_pcie_class, NULL, pm_pcie_cards[numPCICards].pci_dev_num, NULL, PCI_DEV_NAME "%d", numPCICards);

	if(IS_ERR(pm_pcie_cards[numPCICards].device ))
	{
		int err = PTR_ERR(pm_pcie_cards[numPCICards].device);
		pr_info("%s: Error %d while trying to create device %s%d \n", DRV_NAME, err, PCI_DEV_NAME, numPCICards);
		return -1;
	}
	
	printk("%s: Successfully created /dev/%s%d\n", DRV_NAME, PCI_DEV_NAME, numPCICards);

	result = pci_read_config_byte(pdev, 0x34, &capPointer); 


	while(finished == false)
	{
		result = pci_read_config_byte(pdev, capPointer, &capType); 
		
		if(result != 0)
			break;
			
		pr_info("%s: Struct type %x\n", DRV_NAME, capType);
		if(capType == 0x10)
		{
			 finished = true;
		}			
		else
		{
			result = pci_read_config_byte(pdev, capPointer+1, &capPointer); 
		pr_info("%s: NExt offset %x\n", DRV_NAME, capPointer);
		}
			
	} 

	if(finished == true)
	{
		result = pci_read_config_word(pdev, capPointer+0x12, &LinkStatus.ushort);
		pr_info("%s: Link status register 0x%X - Link speed: %d Link width: %d\n", DRV_NAME, LinkStatus.ushort, LinkStatus.bits.CurLinkSpeed, LinkStatus.bits.CurLinkWidth);
		//Save pcie version so we can check it later and switch the clock speed
		pm_pcie_cards[numPCICards].currentConfig.pcieVer = LinkStatus.bits.CurLinkSpeed;
	}
	else
	{
		pm_pcie_cards[numPCICards].currentConfig.pcieVer = 0;
	}
	

	curVal = 1;
	iowrite32(curVal, (ptr1+ REG_IGL2_INT_ENABLED));  

	pm_pcie_cards[numPCICards].pdev = pdev;

	pr_info("%s: IOCTLS \n", DRV_NAME);
	pr_info("%s: IOCTL_IGL2_LEDCTL %lx \n", DRV_NAME, IOCTL_IGL2_LEDCTL);
	pr_info("%s: IOCTL_IGL2_GET_THROUGHPUT %lx \n", DRV_NAME, IOCTL_IGL2_GET_THROUGHPUT);
	pr_info("%s: IOCTL_IGL2_SET_READ_MODE %lx \n", DRV_NAME, IOCTL_IGL2_SET_READ_MODE);
	pr_info("%s: IOCTL_IGL2_SET_ADC_CH %lx \n", DRV_NAME, IOCTL_IGL2_SET_ADC_CH);
	pr_info("%s: IOCTL_IGL2_SET_WRITEMODE %lx \n", DRV_NAME, IOCTL_IGL2_SET_WRITEMODE);
	pr_info("%s: IOCTL_IGL2_CLEAR_DMA_REG %x \n", DRV_NAME, IOCTL_IGL2_CLEAR_DMA_REG);
	pr_info("%s: IOCTL_IGL2_GETCB %lx \n", DRV_NAME, IOCTL_IGL2_GETCB);
	pr_info("%s: IOCTL_IGL2_SET_XFER_SIZE %lx \n", DRV_NAME, IOCTL_IGL2_SET_XFER_SIZE);
	pr_info("%s: IOCTL_IGL2_GET_CONFIG %lx \n", DRV_NAME, IOCTL_IGL2_GET_CONFIG);
	pr_info("%s: IOCTL_IGL2_READ_LSRAM_REG %lx \n", DRV_NAME, IOCTL_IGL2_READ_LSRAM_REG);
	pr_info("%s: IOCTL_IGL2_READ_REG %lx \n", DRV_NAME, IOCTL_IGL2_READ_REG);
	pr_info("%s: IOCTL_IGL2_ACCESS_CONFIG_SPACE %x \n", DRV_NAME, IOCTL_IGL2_ACCESS_CONFIG_SPACE);
	pr_info("%s: IOCTL_IGL2_GET_ADC_STATS %lx \n", DRV_NAME, IOCTL_IGL2_GET_ADC_STATS);
	pr_info("%s: IOCTL_IGL2_SET_VOLT_TOL %lx \n", DRV_NAME, IOCTL_IGL2_SET_VOLT_TOL);
	pr_info("%s: IOCTL_IGL2_RESET_STATS_REG %x \n", DRV_NAME, IOCTL_IGL2_RESET_STATS_REG);
	
	numPCICards++;

	return 0;

err_out_free_irq:
	free_irq(pdev->irq, &pm_pcie_cards[numPCICards]);
	iounmap(pm_pcie_cards[numPCICards].base_addr0);
	iounmap(pm_pcie_cards[numPCICards].base_addr1);
err_out_free_reg:
	release_mem_region(pm_pcie_cards[numPCICards].bar0_base, pm_pcie_cards[numPCICards].bar0_size);
	release_mem_region(pm_pcie_cards[numPCICards].bar1_base, pm_pcie_cards[numPCICards].bar1_size);
err_out_free_res:
	return -ENODEV;
}


static struct pci_driver ms_pci_driver = {
        .name           = DRV_NAME,
        .probe          = ms_pci_probe,
        .remove         = ms_pci_remove,
        .id_table       = ms_pci_tbl,
};

/**
*  pcidrv_init() - Initialize PCIe Module
*
*  Registers Platform Driver
*/

static __init int pcidrv_init(void)
{
	dev_t ms_pci_dev;


	pr_info("%s: Version %s Module init \n", DRV_NAME, DRV_VERSION);
	
	if (alloc_chrdev_region (&ms_pci_dev, 0, MAX_NUM_PCIECARDS, PCI_DEV_NAME) < 0) {
            printk (KERN_ERR "%s: failed to reserve major/minor range\n", DRV_NAME);
            return -1;
    	}

	printk("%s: Got Major %d\n", DRV_NAME, MAJOR(ms_pci_dev) );

	pm_major_id = MAJOR(ms_pci_dev) ;

        pm_pcie_class = class_create (THIS_MODULE, "VIRTUAL");
        
	 pm_pcie_cards = (struct pm_pcie_dev*) kzalloc(MAX_NUM_PCIECARDS * sizeof(struct pm_pcie_dev), GFP_KERNEL);
	if(pm_pcie_cards == NULL)
	{
		return -ENOMEM;
	}

	if(pci_register_driver(&ms_pci_driver))
	{
		pr_info("%s: unable to register pci driver \n", DRV_NAME);
		return -1;
	}

	return 0;
}

/**
* pcidrv_cleanup() - Uninitialize the PCIe Module
*
*  Unregister the Platform Driver
*/

static __exit void  pcidrv_cleanup(void)
{	
	int count = 0;
	int curVal = 0;
	dev_t ms_pci_dev = MKDEV(pm_major_id, 0);
	unsigned int* ptr1 = NULL;
	
	pr_info("%s: Module being removed \n", DRV_NAME);
	
	for( count = 0; count < numPCICards && count < MAX_NUM_PCIECARDS; count++)
	{
		pr_info("%s: %d Stop interrupts \n", DRV_NAME, count);
		ptr1 = (unsigned int *)(pm_pcie_cards[count].base_addr1);
		iowrite32(curVal, (ptr1 + REG_IGL2_INT_ENABLED));  		

		free_irq(pm_pcie_cards[count].pdev->irq, (void*)&pm_pcie_cards[count]);

		pr_info("%s: %d Disable device \n", DRV_NAME, count);
		pci_disable_msi(pm_pcie_cards[count].pdev);
 		pci_clear_master(pm_pcie_cards[count].pdev);
		pci_disable_device(pm_pcie_cards[count].pdev);

		pr_info("%s: %d Clean up memory \n", DRV_NAME, count);
		iounmap(pm_pcie_cards[count].base_addr0);
		iounmap(pm_pcie_cards[count].base_addr1);
		release_mem_region(pm_pcie_cards[count].bar0_base, pm_pcie_cards[count].bar0_size);
		release_mem_region(pm_pcie_cards[count].bar1_base, pm_pcie_cards[count].bar1_size);

		dma_free_coherent(NULL, MAX_DMA_BUFFER, pm_pcie_cards[count].addr_wr, pm_pcie_cards[count].bus_addr_wr);
		dma_free_coherent(NULL, MAX_DMA_BUFFER, pm_pcie_cards[count].addr_rd, pm_pcie_cards[count].bus_addr_rd);

 		cdev_del( pm_pcie_cards[count].pci_cdev);

	 	device_destroy (pm_pcie_class, pm_pcie_cards[count].pci_dev_num);
	}

	kfree((const void *)pm_pcie_cards);

	pr_info("%s: Unregister driver \n", DRV_NAME);
	pci_unregister_driver (&ms_pci_driver);

	pr_info("%s: Class cleanup \n", DRV_NAME);

        class_destroy (pm_pcie_class);
	unregister_chrdev_region(ms_pci_dev, MAX_NUM_PCIECARDS);
	printk(KERN_INFO "%s: Driver unregistered \n", DRV_NAME);
}

module_init(pcidrv_init);
module_exit(pcidrv_cleanup);

MODULE_AUTHOR("PassMark");
MODULE_DESCRIPTION("PCI device driver");
MODULE_LICENSE("GPL");
